__author__ = "Dan"


class AvatarAppearance:
    """
    The presentation aspect of an avatar.
    """

    def __init__(self, body_stroke, body_fill, eye_stroke, eye_fill):
        self.body_stroke = body_stroke
        self.body_fill = body_fill
        self.eye_stroke = eye_stroke
        self.eye_fill = eye_fill



import logging
from typing import TYPE_CHECKING, List

from simulation.action import ACTIONS, MoveAction, WaitAction
from simulation.direction import Direction

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    from simulation.interactables.pickups.artefacts import _Artefact


class AvatarWrapper(object):
    """
    The application's view of a character, not to be confused with "Avatar",
    the player-supplied code.
    """

    def __init__(self, player_id, initial_location, avatar_appearance):
        self.player_id = player_id
        self.location = initial_location
        self.previous_location = initial_location
        self.orientation = "north"
        self.health = 5
        self.score = 0
        self.backpack: "List[_Artefact]" = []
        self.BACKPACK_SIZE = 10
        self.events = []
        self.avatar_appearance = avatar_appearance
        self.effects = set()
        self.resistance = 0
        self.attack_strength = 1
        self.fog_of_war_modifier = 0
        self.logs = []
        self._action = None

    def update_effects(self):
        effects_to_remove = set()
        for effect in self.effects:
            effect.on_turn()
            if effect.is_expired:
                effects_to_remove.add(effect)
        for effect in effects_to_remove:
            effect.remove()

    @property
    def action(self):
        return self._action

    @property
    def is_moving(self):
        return isinstance(self.action, MoveAction)

    def _construct_action(self, action_data):
        action_type = action_data["action_type"]
        action_args = action_data.get("options", {})
        action_args["avatar"] = self
        return ACTIONS[action_type](**action_args)

    def calculate_orientation(self):
        """
        Calculates the orientation of the avatar (ie. what direction the avatar is pointed
        towards) for rendering in the front end of the game.
        :return: A string representation of a cardinal direction.
        """
        _current_location = self.location
        _previous_location = self.previous_location

        direction_of_orientation = Direction(
            _current_location.x - _previous_location.x,
            _current_location.y - _previous_location.y,
        )

        if _current_location == _previous_location:
            return self.orientation

        return direction_of_orientation.cardinal

    def decide_action(self, serialized_action):
        try:
            action = self._construct_action(serialized_action)

        except (KeyError, ValueError) as err:
            LOGGER.error("Bad action data supplied: %s", err)
        except TypeError as err:
            LOGGER.error("Worker data not received: %s", err)
        else:
            self._action = action
            return True

        # Returning False here means that the action won't be registered on the cell
        # (although it will be registered on the avatar)
        self._action = WaitAction(self)
        return False

    def clear_action(self):
        self._action = None

    def clear_logs(self):
        self.logs = []

    def die(self, respawn_location):
        # TODO: extract settings for health and score loss on death
        self.health = 5
        self.score = max(0, self.score - 2)
        self.location = respawn_location

    def add_event(self, event):
        self.events.append(event)

    def damage(self, amount):
        applied_dmg = max(0, amount - self.resistance)
        self.health -= applied_dmg
        return applied_dmg

    def backpack_has_space(self):
        return len(self.backpack) < self.BACKPACK_SIZE

    def serialize(self):
        return {
            "health": self.health,
            "location": self.location.serialize(),
            "score": self.score,
            "id": self.player_id,
            "orientation": self.orientation,
            "backpack": [artefact.serialize() for artefact in self.backpack],
        }

    def __repr__(self):
        return f"Avatar(id={self.player_id}, location={self.location}, health={self.health}, score={self.score}, backpack={[artefact.serialize() for artefact in self.backpack]})"



"""
Sets up the AimmoAppConfig class. Since we don't want to reset the app config every
time the Django server reloads when we make a code change, the app config is set only
if the main thread isn't already running.
"""
import os

if os.environ.get("RUN_MAIN", None) != "true":
    default_app_config = "aimmo.apps.AimmoAppConfig"

__version__ = '0.69.10b450'



# -*- coding: utf-8 -*-
# Generated by Django 1.10.8 on 2019-04-18 07:28
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("aimmo", "0009_add_game_status")]

    operations = [
        migrations.AlterField(
            model_name="game",
            name="auth_token",
            field=models.CharField(blank=True, max_length=24),
        )
    ]



import aiohttp
import os
from asyncio import TimeoutError


class DjangoCommunicator(object):
    """
    This class encapsulates the communication between aimmo-game
    and the django server
    """

    def __init__(self, django_api_url):
        timeout = aiohttp.ClientTimeout(total=1)
        self.session = aiohttp.ClientSession(timeout=timeout)
        self.django_api_url = django_api_url

    async def get_game_metadata(self):
        try:
            async with self.session.get(f"{self.django_api_url}users/") as response:
                if response.status == 200:
                    return await response.json()
                else:
                    raise GameMetadataFetchFailedError
        except (aiohttp.ClientConnectionError, aiohttp.ContentTypeError, TimeoutError):
            raise GameMetadataFetchFailedError

    async def patch_token(self, data):
        response = await self.session.patch(
            f"{self.django_api_url}token/",
            headers={"Game-token": os.environ["TOKEN"]},
            json=data,
        )
        return response

    async def patch_game(self, data):
        response = await self.session.patch(
            self.django_api_url, headers={"Game-token": os.environ["TOKEN"]}, json=data
        )
        return response

    async def close_session(self, app):
        await self.session.close()


class GameMetadataFetchFailedError(Exception):
    pass



from os import urandom

from base64 import urlsafe_b64encode
from common.models import Class
from django.contrib.auth.models import User
from django.db import models
from wagtail.snippets.models import register_snippet

from aimmo import app_settings

DEFAULT_WORKSHEET_ID = 1

GAME_GENERATORS = [("Main", "Open World")] + [  # Default
    ("Level%s" % i, "Level %s" % i) for i in range(1, app_settings.MAX_LEVEL + 1)
]


def generate_auth_token():
    return urlsafe_b64encode(urandom(16))


class WorksheetManager(models.Manager):
    def sorted(self):
        return self.get_queryset().order_by("sort_order")


@register_snippet
class Worksheet(models.Model):
    ERA_CHOICES = [
        (1, "future"),
        (2, "ancient"),
        (3, "modern day"),
        (4, "prehistoric"),
        (5, "broken future"),
    ]

    name = models.CharField(max_length=100)
    description = models.TextField(blank=True)
    short_description = models.TextField(blank=True)
    image_path = models.CharField(max_length=255, blank=True)
    active_image_path = models.CharField(max_length=255, blank=True)
    era = models.PositiveSmallIntegerField(
        choices=ERA_CHOICES, default=ERA_CHOICES[0][0]
    )
    thumbnail_text = models.CharField(max_length=100, blank=True)
    thumbnail_image_path = models.CharField(max_length=255, blank=True)
    teacher_pdf_name = models.CharField(max_length=255, blank=True)
    student_pdf_name = models.CharField(max_length=255, blank=True)
    sort_order = models.IntegerField(default=0)

    objects = WorksheetManager()

    starter_code = models.TextField()

    def __str__(self):
        return f"{self.id}: {self.name}"


class Game(models.Model):
    RUNNING = "r"
    STOPPED = "s"
    PAUSED = "p"
    STATUS_CHOICES = ((RUNNING, "running"), (STOPPED, "stopped"), (PAUSED, "paused"))

    name = models.CharField(max_length=100, blank=True, null=True)
    auth_token = models.CharField(max_length=24, blank=True)
    owner = models.ForeignKey(
        User,
        blank=True,
        null=True,
        related_name="owned_games",
        on_delete=models.SET_NULL,
    )
    game_class = models.OneToOneField(
        Class,
        verbose_name="Class",
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    public = models.BooleanField(default=False)
    can_play = models.ManyToManyField(
        User,
        related_name="playable_games",
        help_text="List of auth_user IDs of users who are allowed to play and have access to the game.",
    )
    completed = models.BooleanField(default=False)
    main_user = models.ForeignKey(
        User,
        blank=True,
        null=True,
        related_name="games_for_user",
        on_delete=models.SET_NULL,
    )
    static_data = models.TextField(blank=True, null=True)

    # Game config
    generator = models.CharField(
        max_length=20, choices=GAME_GENERATORS, default=GAME_GENERATORS[0][0]
    )
    target_num_cells_per_avatar = models.FloatField(default=16)
    target_num_score_locations_per_avatar = models.FloatField(default=0.5)
    score_despawn_chance = models.FloatField(default=0.05)
    target_num_pickups_per_avatar = models.FloatField(default=1.2)
    pickup_spawn_chance = models.FloatField(default=0.1)
    obstacle_ratio = models.FloatField(default=0.1)
    start_height = models.IntegerField(default=31)
    start_width = models.IntegerField(default=31)
    status = models.CharField(max_length=1, choices=STATUS_CHOICES, default=RUNNING)
    worksheet = models.ForeignKey(
        Worksheet, on_delete=models.PROTECT, default=DEFAULT_WORKSHEET_ID
    )

    @property
    def is_active(self):
        return not self.completed

    def __str__(self):
        return self.name

    def can_user_play(self, user: User) -> bool:
        """Checks whether the given user has permission to play the game.

        A user can play the game if they are part of the game's class or
        the teacher of that class.

        Args:
            user: A standard django User object

        Returns:
            bool: True if user can play the game, False otherwise
        """
        try:
            return (
                self.game_class.students.filter(new_user=user).exists()
                or user == self.game_class.teacher.new_user
            )
        except AttributeError:
            return False

    def settings_as_dict(self):
        return {
            "GENERATOR": self.generator,
            "OBSTACLE_RATIO": self.obstacle_ratio,
            "PICKUP_SPAWN_CHANCE": self.pickup_spawn_chance,
            "SCORE_DESPAWN_CHANCE": self.score_despawn_chance,
            "START_HEIGHT": self.start_height,
            "START_WIDTH": self.start_width,
            "TARGET_NUM_CELLS_PER_AVATAR": self.target_num_cells_per_avatar,
            "TARGET_NUM_PICKUPS_PER_AVATAR": self.target_num_pickups_per_avatar,
            "TARGET_NUM_SCORE_LOCATIONS_PER_AVATAR": self.target_num_score_locations_per_avatar,
        }

    def save(self, *args, **kwargs):
        super(Game, self).full_clean()
        super(Game, self).save(*args, **kwargs)


class Avatar(models.Model):
    owner = models.ForeignKey(User, on_delete=models.CASCADE)
    game = models.ForeignKey(Game, on_delete=models.CASCADE)
    code = models.TextField()
    auth_token = models.CharField(max_length=24, default=generate_auth_token)

    class Meta:
        unique_together = ("owner", "game")



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("aimmo", "0001_initial")]

    operations = [
        migrations.RemoveField(model_name="avatar", name="player"),
        migrations.DeleteModel(name="Avatar"),
    ]



def next_turn(world_state, avatar_state):
    world_state = world_state
    avatar_state = avatar_state

    if world_state.get_cell(avatar_state.location).generates_score:
        return WaitAction()

    possible_directions = get_possible_directions()
    directions_to_emphasise = [
        d for d in possible_directions if is_towards(d, get_closest_score_location())
    ]
    return MoveAction(
        random.choice(possible_directions + (directions_to_emphasise * 5))
    )


def is_towards(direction, location):
    if location:
        return distance_between(avatar_state.location, location) > distance_between(
            avatar_state.location + direction, location
        )
    else:
        return False


def distance_between(a, b):
    return abs(a.x - b.x) + abs(a.y - b.y)


def get_closest_score_location(self):
    score_cells = list(world_state.score_cells())
    if score_cells:
        return min(
            score_cells,
            key=lambda cell: distance_between(cell.location, avatar_state.location),
        ).location
    else:
        return None


def get_possible_directions():
    directions = (direction.EAST, direction.SOUTH, direction.WEST, direction.NORTH)
    return [d for d in directions if world_state.can_move_to(avatar_state.location + d)]



from simulation.interactables.pickups.artefacts import (
    ChestArtefact,
    KeyArtefact,
    YellowOrbArtefact,
)
from simulation.interactables.pickups.damage_boost_pickup import DamageBoostPickup
from simulation.interactables.pickups.health_pickup import HealthPickup
from simulation.interactables.pickups.invulnerability_pickup import (
    InvulnerabilityPickup,
)


def serialize_pickups(world_map):
    return [cell.interactable.serialize() for cell in world_map.pickup_cells()]


ALL_PICKUPS = (
    DamageBoostPickup,
    InvulnerabilityPickup,
    HealthPickup,
    YellowOrbArtefact,
    ChestArtefact,
    KeyArtefact,
)



def next_turn(world_state, avatar_state):
    avatar_state = avatar_state
    location = avatar_state.location

    directions = (direction.EAST, direction.SOUTH, direction.WEST, direction.NORTH)
    direction_of_other_avatar = next(
        (
            d
            for d in directions
            if world_state.is_visible(location + d)
            and world_state.get_cell(location + d).avatar
        ),
        None,
    )
    if direction_of_other_avatar:
        return AttackAction(direction_of_other_avatar)

    direction_to_other_player = direction_to(
        next(
            cell.location
            for cell in world_state.all_cells()
            if cell.avatar and cell.location != avatar_state.location
        )
    )
    if direction_to_other_player:
        return MoveAction(
            random.choice(directions + ((direction_to_other_player,) * 10))
        )
    return MoveAction(random.choice(directions))


def direction_to(location):
    vector_to = location - location
    if vector_to.x != 0:
        return direction.Direction(1 if vector_to.x > 0 else -1, 0)
    if vector_to.y != 0:
        return direction.Direction(0, 1 if vector_to.y > 0 else -1)
    return None



import json
from rest_framework import serializers

from aimmo.models import Game, Avatar, Worksheet
from aimmo.game_manager import GameManager


class GameSerializer(serializers.Serializer):
    name = serializers.CharField(max_length=100, required=False)
    settings = serializers.SerializerMethodField("get_settings_as_dict")
    status = serializers.CharField(max_length=1, required=False)
    class_id = serializers.SerializerMethodField()
    worksheet_id = serializers.CharField(max_length=1, required=False)
    era = serializers.SerializerMethodField("get_worksheet_era")

    def get_class_id(self, game: Game):
        try:
            return str(game.game_class.id)
        except AttributeError:
            return None

    def get_worksheet_id(self, game: Game):
        try:
            return str(game.worksheet.id)
        except AttributeError:
            return "2"

    def get_worksheet_era(self, game: Game):
        try:
            return str(game.worksheet.era)
        except AttributeError:
            return "1"

    def get_settings_as_dict(self, game: Game):
        return json.dumps(game.settings_as_dict(), sort_keys=True)

    def update(self, instance, validated_data):
        instance.name = validated_data.get("name", instance.name)
        instance.status = validated_data.get("status", instance.status)
        instance.worksheet_id = validated_data.get(
            "worksheet_id", instance.worksheet_id
        )

        instance.save()

        if "worksheet_id" in validated_data:
            avatars = Avatar.objects.filter(game=instance)
            worksheet = Worksheet.objects.get(id=instance.worksheet_id)

            for avatar in avatars:
                avatar.code = worksheet.starter_code
                avatar.save()

            # If the game is running, the game server needs to be restarted
            if instance.status == Game.RUNNING:
                game_manager = GameManager()
                game_manager.recreate_game_server(
                    game_id=instance.id,
                    game_data_updates={"worksheet_id": str(instance.worksheet_id)},
                )

        return instance


class GameIdsSerializer(serializers.Serializer):
    game_ids = serializers.MultipleChoiceField(choices=[])



from __future__ import annotations

from typing import TYPE_CHECKING

from simulation.interactables import serialize_interactables
from simulation.worksheet.worksheet import get_worksheet_data

if TYPE_CHECKING:
    from simulation.world_map import WorldMap
    from simulation.avatar.avatar_manager import AvatarManager
    from simulation.worksheet.worksheet import WorksheetData


class GameState:
    """
    Encapsulates the entire game state, including avatars, their code, and the world.
    """

    def __init__(self, world_map, avatar_manager, worksheet: WorksheetData = None):
        if worksheet is None:
            worksheet = get_worksheet_data()

        self.world_map: WorldMap = world_map
        self.avatar_manager: AvatarManager = avatar_manager
        self.turn_count: int = 0
        self.worksheet: WorksheetData = worksheet

    def serialize(self):
        return {
            "era": self.worksheet.era,
            "southWestCorner": self.world_map.get_serialized_south_west_corner(),
            "northEastCorner": self.world_map.get_serialized_north_east_corner(),
            "players": self.avatar_manager.serialize_players(),
            "interactables": serialize_interactables(self.world_map),
            "obstacles": self.world_map.serialize_obstacles(),
            "turnCount": self.turn_count,
        }



# Generated by Django 2.0 on 2020-10-28 17:30

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('aimmo', '0022_allow_games_to_have_no_name'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='levelattempt',
            unique_together=set(),
        ),
        migrations.RemoveField(
            model_name='levelattempt',
            name='game',
        ),
        migrations.RemoveField(
            model_name='levelattempt',
            name='user',
        ),
        migrations.DeleteModel(
            name='LevelAttempt',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2018-08-28 10:15
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("aimmo", "0007_auto_20180515_1300")]

    operations = [
        migrations.AlterField(
            model_name="game", name="public", field=models.BooleanField(default=False)
        )
    ]



class LogCollector:
    """
    This class aggregates:
    - the avatar logs (logs outputted by the game under certain conditions)
    These logs are concatenated to form the `playerLogs`.
    """

    def __init__(self, avatar_manager):
        super(LogCollector, self).__init__()

        self.avatar_manager = avatar_manager

    def collect_logs(self, user_id):
        avatar = self.avatar_manager.get_avatar(user_id)

        return "\n".join(avatar.logs)



"""
Any helper functions used for the game.
"""
from __future__ import absolute_import

from django.shortcuts import get_object_or_404, render

from aimmo import app_settings, exceptions

from .models import Game


def render_game(request, game):
    """
    :param request: Request object used to generate this response.
    :param game: Game object.
    :return: HttpResponse object with a given context dictionary and a template.
    """
    context = {
        "current_user_player_key": request.user.pk,
        "active": game.is_active,
        "static_data": game.static_data or "{}",
    }

    return render(request, "players/game_ide.html", context)


def get_environment_connection_settings(game_id):
    """
    This function will return the correct URL parts and a SSL flag
    based on the environment of which it exists in.

    :param game_id: Integer with the ID of the game.
    :return: A dict object with all relevant settings.
    """
    game_url_base_and_path = app_settings.GAME_SERVER_URL_FUNCTION(game_id)
    return {
        "game_url_base": game_url_base_and_path[0],
        "game_url_path": game_url_base_and_path[1],
        "game_ssl_flag": app_settings.GAME_SERVER_SSL_FLAG,
        "game_id": game_id,
    }


def get_avatar_id_from_user(user, game_id):
    """
    A helper function which will return an avatar ID that is assigned to a
    particular django user owner in a game.
    :param user: A django user object taken from the request.
    :param game_id: The game ID in which the avatar is being requested from.
    :return: An integer containing the avatar_ID.
    """

    game = get_object_or_404(Game, id=game_id)
    if not game.can_user_play(user):
        raise exceptions.UserCannotPlayGameException

    avatar = game.avatar_set.get(owner=user.id)

    return avatar.id
